/******************************************************************************
 *
 *    File Name:  addr_latch.v
 *      Version:  2.5
 *         Date:  June 05, 2000
 * 
 *  Description:  generates row, column, and ba addresses for DDR SDRAM
 *                also generates cas_lat_max and burst_8, burst_2
 *
 *       Author:  Jennnifer Tran
 *      Company:  Xilinx
 *
 *                Copyright (c) 1999 Xilinx, Inc.
 *                All rights reserved
 *
 *    DDR SDRAM:  64M (1M x 16Bit x 4Bank)
 * 
 *   Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
 *                WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
 *                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
 *                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
 *
 ******************************************************************************/
`timescale 100ps / 10ps
`include "../src/define.v"

module addr_latch(/*AUTOARG*/
   // Outputs
   ddr_ad, cas_lat_max, cas_lat_half, ddr_ba, burst_2, burst_8, 
   // Inputs
   u_addr, row_addr, clk, reset1, reset2, mrs_addr
   );
   
   output [`DDR_ADDR_MSB:0] ddr_ad;
   output [1:0] 	    cas_lat_max;
   output [1:0]		    cas_lat_half;
   output [1:0] 	    ddr_ba;
   output 		    burst_2, burst_8;
   
   input [`U_ADDR_MSB:0]    u_addr;
   input 		    row_addr;
   input 		    clk;
   input 		    reset1, reset2;
   input 		    mrs_addr;
   
   
   reg [1:0] 		    ddr_ba_o, ddr_ba_o_P1;
   reg [`DDR_ADDR_MSB:0]    ddr_ad_o, ddr_ad_o_P1;
   reg [1:0] 		    next_cas_lat_max, cas_lat_max;
   reg 			    next_cas_lat_half;
   reg [1:0] 		    cas_lat_half;
   wire [1:0] 		    ddr_burst_length = u_addr[1:0];
   wire [2:0] 		    ddr_cas_latency = u_addr[6:4];
   
   reg 			    burst_2, burst_8;
   reg 			    next_burst_2, next_burst_8;
   
   //decode burst length
   always @(ddr_burst_length) begin
      case (ddr_burst_length) 
	  2'b01:   begin next_burst_2 = 1; next_burst_8 = 0; end
	  2'b10:   begin next_burst_2 = 0; next_burst_8 = 0; end 
	  default: begin next_burst_2 = 0; next_burst_8 = 1; end 
      endcase 
   end // always @ (ddr_burst_length)
   
   //generate cas_latency_max, cas_lat_half
   always @(ddr_cas_latency)
     case (ddr_cas_latency) 
       3'b011:  begin
	  next_cas_lat_max = 2'b10; //cas_lat = 3
	  next_cas_lat_half = 1'b0;
       end
       3'b101:  begin
	  next_cas_lat_max = 2'b00; //cas_lat = 1.5
	  next_cas_lat_half = 1'b1;
       end
       3'b110:  begin
	  next_cas_lat_max = 2'b01; //cas_lat = 2.5
	  next_cas_lat_half = 1'b1;
       end
       default: begin
	  next_cas_lat_max = 2'b01; //cas_lat = 2
	  next_cas_lat_half = 1'b0;
       end
     endcase 

   //generate address & ba
   //address & ba are generated on negative edge of clk to meet DDR hold time
   //ba is always on u_addr[21:20]
   //Mode register values are on u_addr[8:0]
   //for READ/WRITE:row addr are on u_addr[19:8]
   //               col addr are on u_addr[7:0], A10=1:auto_precharge
   //for precharge: A10=1:all banks (for mt46v2m32, A8 needs to be 1)
   always @(negedge clk) begin
      ddr_ba_o <= u_addr[21:20];
      if (mrs_addr)
	ddr_ad_o <= {3'b000,u_addr[8:0]};  
      else if (row_addr)
	ddr_ad_o <= u_addr[19:8];           //row addr
      else
	ddr_ad_o <= {4'b0101,u_addr[7:0]};    //col addr, 
   end
   always @(negedge clk) begin
      ddr_ad_o_P1 <= ddr_ad_o;	 
      ddr_ba_o_P1 <= ddr_ba_o;
   end
        
   always @(posedge clk) begin
      if (reset2) begin
	 cas_lat_max <= 3'b00;
      end
      else begin 
	 if (mrs_addr) begin
	    cas_lat_max <= next_cas_lat_max;
	    cas_lat_half <= {2{next_cas_lat_half}};
	    burst_2 <= next_burst_2;
	    burst_8 <= next_burst_8;
	 end
      end
   end
   
   //add IO buffers to get out of FPGA
   OBUF_SSTL2_I ddr_ba0  (.O(ddr_ba[0]),  .I(ddr_ba_o_P1[0] ));
   OBUF_SSTL2_I ddr_ba1  (.O(ddr_ba[1]),  .I(ddr_ba_o_P1[1] ));
   
   wire [11:0] ddr_ad_o_dly = ddr_ad_o_P1;  
   
`include "../src/ddr_addr_io_12.v"
endmodule // addr_latch






