--*****************************************************************************
--
--    File Name:  addr_latch.vhd
--      Version:  2.5
--         Date:  June 05, 2000
-- 
--  Description:  generates row, column, and ba addresses for DDR SDRAM
--                also generates cas_lat_max, burst_8, bust_4, and burst_2
--
--       Author:  Jennnifer Tran
--      Company:  Xilinx
--
--                Copyright (c) 1999 Xilinx, Inc.
--                All rights reserved
--
--    DDR SDRAM:  64M (1M x 16Bit x 4Bank)
-- 
--   Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
--                WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--*****************************************************************************/


library IEEE;
use IEEE.std_logic_1164.all;
library work;
use WORK.ddr_controller_comp.all;
-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

entity addr_latch is
  port(
    --Inputs
    u_addr	  : in std_logic_vector(U_ADDR_MSB downto 0);
    row_addr	  : in std_logic;
    mrs_addr	  : in std_logic;
    clk           : in std_logic;
    reset1        : in std_logic;
    reset2        : in std_logic;
    
    --Outputs
    burst_8	  : out std_logic;
    burst_2	  : out std_logic;
    cas_lat_max   : out std_logic_vector(1 downto 0);
    cas_lat_half  : out std_logic_vector(1 downto 0);
    ddr_ad	  : out std_logic_vector(DDR_ADDR_MSB downto 0);
    ddr_ba	  : out std_logic_vector(1 downto 0));
end addr_latch;

architecture rtl of addr_latch is
  
  signal ddr_ba_o         	: std_logic_vector(1 downto 0);
  signal ddr_ba_o_P1      	: std_logic_vector(1 downto 0);
  signal ddr_ad_o         	: std_logic_vector(DDR_ADDR_MSB downto 0);
  signal ddr_ad_o_P1      	: std_logic_vector(DDR_ADDR_MSB downto 0);
  signal next_cas_lat_max 	: std_logic_vector(1 downto 0);
  signal ddr_burst_length 	: std_logic_vector(1 downto 0);
  signal ddr_cas_latency  	: std_logic_vector(2 downto 0);
  signal next_cas_lat_half      : std_logic;
  signal next_burst_2           : std_logic;
  signal next_burst_8           : std_logic;
  
  signal ddr_ad_o_dly           : std_logic_vector(DDR_ADDR_MSB downto 0);

  component OBUF_SSTL2_I
    port(I : in std_logic;
         O : out std_logic);
  end component;

begin

  ddr_burst_length(1 downto 0) <= u_addr(1 downto 0);
  ddr_cas_latency(2 downto 0)  <= u_addr(6 downto 4);
  
  --decode burst length
  process (ddr_burst_length)
  begin
    case ddr_burst_length is
      when "01"    => next_burst_2 <= '1'; next_burst_8 <= '0';
      when "10"    => next_burst_2 <= '0'; next_burst_8 <= '0';
      when others  => next_burst_2 <= '0'; next_burst_8 <= '1';
    end case;
  end process;

  --generate cas_latency_max, cas_lat_half
  process (ddr_cas_latency)
  begin
    case ddr_cas_latency is
      when "011"  => next_cas_lat_max <="10"; next_cas_lat_half <= '0'; --cas_lat = 3,   
      when "101"  => next_cas_lat_max <="00"; next_cas_lat_half <= '1'; --cas_lat = 1.5,   
      when "110"  => next_cas_lat_max <="01"; next_cas_lat_half <= '1'; --cas_lat = 2.5,   
      when others => next_cas_lat_max <="01"; next_cas_lat_half <= '0'; --cas_lat = 2,   
    end case;
  end process;
  
  --generate address & ba
  --address & ba are generated on negative edge of clk to meet DDR hold time
  --ba is always on u_addr[21:20]
  --Mode register values are on u_addr[8:0]
  --for READ/WRITE:row addr are on u_addr[19:8]
  --               col addr are on u_addr[7:0], A10=1:auto_precharge
  --for precharge: A10=1:all banks
  
  process (clk)
  begin
    if (clk 'event and clk = '0') then
      ddr_ba_o <= u_addr(21 downto 20);
      if (mrs_addr = '1') then
        ddr_ad_o   <= "000" & u_addr(8 downto 0);
      elsif (row_addr = '1') then
        ddr_ad_o   <= u_addr(19 downto 8);
      else
        ddr_ad_o   <= "0100" & u_addr(7 downto 0);
      end if;
    end if;
  end process;

  process (clk, reset1)
  begin
    if (reset1 = '1') then
      ddr_ad_o_P1 <= (others=>'0');
    elsif (clk 'event and clk = '0') then
      ddr_ad_o_P1 <= ddr_ad_o;
    end if;
  end process;
        
  process (clk, reset2)
  begin
    if (reset2 = '1') then
      ddr_ba_o_P1 <= "00";
    elsif (clk 'event and clk = '0') then
      ddr_ba_o_P1 <= ddr_ba_o;
    end if;
  end process;

  process (clk, reset2)
  begin
    if (reset2 = '1') then
      cas_lat_max <= "00";
    elsif (clk 'event and clk = '1') then
      if mrs_addr = '1' then
        cas_lat_max <= next_cas_lat_max;
        cas_lat_half <= next_cas_lat_half & next_cas_lat_half;
        burst_2 <= next_burst_2;
        burst_8 <= next_burst_8;
      end if;      
    end if;
  end process;
        
  ddr_ad_o_dly <= ddr_ad_o_P1 after 1 ns;
  -- the 1.5 ns is to simulate clk-to-out delay.
  -- This eliminate setup & hold error in functional simulation.
  
  g0: for i in 0 to 11 generate
    o_ddr_ad: OBUF_SSTL2_I port map (I=>ddr_ad_o_dly(i),O=>ddr_ad(i));
  end generate;

  o_ddr_ba0: OBUF_SSTL2_I port map (I=>ddr_ba_o_P1(0),O=>ddr_ba(0));
  o_ddr_ba1: OBUF_SSTL2_I port map (I=>ddr_ba_o_P1(1),O=>ddr_ba(1));

end rtl;

