--*****************************************************************************
--
--    File Name:  controller.vhd
--      Version:  2.5
--         Date:  June 05, 2000
-- 
--  Description:  state machine to generate all control signals for DDR SDRAM 
--                and other blocks of the DDR controller
-- 
--                state machine is implemented as one-hot for performance
--                all outputs are registered
--                
--                ddr_dqs_t is generated by negative edge of clk2x
--                All control signals are generated by positive edge of clk, 
-- 
-- 
--       Author:  Jennnifer Tran
--      Company:  Xilinx
--
--                Copyright (c) 1999 Xilinx, Inc.
--                All rights reserved
--
--    DDR SDRAM:  64M (1M x 16Bit x 4Bank)
-- 
--   Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
--                WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--*****************************************************************************/

library IEEE;
use IEEE.std_logic_1164.all;
library work;
use WORK.ddr_controller_comp.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on


entity controller is
  port(
    --Outputs
    ddr_rasb		: out std_logic;
    ddr_casb		: out std_logic;
    ddr_web		: out std_logic;
    ld_burst            : out std_logic;
    ld_rcd		: out std_logic;
    ld_cas_lat		: out std_logic;
    u_ref_ack		: out std_logic;
    ddr_read_en		: out std_logic_vector(ENABLE_MSB downto 0);
    ddr_write_en	: out std_logic_vector(ENABLE_MSB downto 0);
    ddr_dqs_toggle	: out std_logic;
    u_data_valid_en	: out std_logic;
    ddr_dqs_t           : out std_logic;
    mrs_addr		: out std_logic;
    row_addr		: out std_logic;
    
    --Inputs
    u_cmd		: in std_logic_vector(7 downto 1);
    cntlr_reset1        : in std_logic;
    cntlr_reset2        : in std_logic;
    cntlr_reset3        : in std_logic;
    clk                 : in std_logic;
    clk2x		: in std_logic;
    lac_clk_in		: in std_logic;
    burst_end		: in std_logic;
    cas_lat_end		: in std_logic;
    rcd_end		: in std_logic;
    burst_8		: in std_logic;
    burst_2		: in std_logic);
end controller;

architecture rtl of controller is
  
  signal ddr_cmd, next_ddr_cmd  : std_logic_vector(2 downto 0);
  signal state, next_state      : std_logic_vector(12 downto 1);   
  signal ddr_rasb_o        	: std_logic;
  signal ddr_rasb_o_P1     	: std_logic;
  signal ddr_casb_o        	: std_logic;
  signal ddr_casb_o_P1     	: std_logic;
  signal ddr_web_o         	: std_logic;
  signal ddr_web_o_P1      	: std_logic;
  signal data_read         	: std_logic;
  signal data_write        	: std_logic;
  signal data_read_write   	: std_logic;
  signal lac_clk_in_inv	   	: std_logic;
  signal logic1	           	: std_logic;
  signal logic0	           	: std_logic;
  signal read_cmd          	: std_logic;
  signal read_cmd_reg      	: std_logic;
  signal clk2x_inv         	: std_logic;
  signal next_mrs_addr	   	: std_logic;
  signal next_row_addr	   	: std_logic;
  signal next_ddr_read_en  	: std_logic;
  signal next_ld_cas_lat   	: std_logic;
  signal next_ld_burst	   	: std_logic;
  signal next_ld_rcd	   	: std_logic;
  signal next_u_ref_ack	   	: std_logic;
  signal next_ddr_rasb_o   	: std_logic;
  signal next_ddr_casb_o   	: std_logic;
  signal next_ddr_web_o	   	: std_logic;
  signal next_ddr_write_en 	: std_logic;
  signal next_ddr_write_en_P1 	: std_logic;    
  signal next_u_data_valid_en 	: std_logic;                         	
  signal ddr_dqs_t_tmp1	   	: std_logic;
  signal ddr_dqs_t_tmp2	   	: std_logic;
  signal lac_clk	   	: std_logic;
  signal lac_clk_out	   	: std_logic;
  signal lac_clk_wire	   	: std_logic;
  signal next_ddr_read_en_P1    : std_logic;
  
  component OBUF_SSTL2_I
    port(I : in std_logic;
         O : out std_logic);
  end component;
  component FD
    port(D  : in std_logic;
         C  : in std_logic;
         Q  : out std_logic);
  end component;

begin

  data_read<=state(CTLR_READ_DATA);
  data_write<=state(CTLR_WRITE_DATA);
  data_read_write<=data_read or data_write;
  logic1 <= '1';
  logic0 <= '0';

  -----------------------------------------------------------------------------
  -- state machine ------------------------------------------------------------
  -----------------------------------------------------------------------------
  process(read_cmd_reg,burst_8,burst_2,u_cmd,state,burst_end,cas_lat_end,rcd_end)
  begin
    --has default values for outputs so synthesis tool don't infer latches
    next_state <= (others=>'0');
    read_cmd <= '0';
        
    if state(CTLR_IDLE) = '1' then
      if u_cmd(SYS_REFRESH) = '1' then
        next_state(CTLR_REFRESH)<=logic1;
      elsif u_cmd(SYS_PRECHARGE) ='1' then
        next_state(CTLR_PRECHARGE)<=logic1;
      elsif u_cmd(SYS_READ) = '1' then
        next_state(CTLR_ACT)<=logic1;
        read_cmd <= '1';
      elsif u_cmd(SYS_WRITE) ='1' then
        next_state(CTLR_ACT)<=logic1;
      elsif u_cmd(SYS_LOAD_MR) = '1' then
        next_state(CTLR_LOAD_MR)<=logic1;
      else
        next_state(CTLR_IDLE)<=logic1;
      end if;
      
    elsif state(CTLR_ACT) = '1' then    -- issue ACTIVE command
      if (rcd_end = '1') then
        if (read_cmd_reg = '1') then
          next_state(CTLR_READ)<=logic1;
          read_cmd <= '1';
        else
          next_state(CTLR_WRITE)<=logic1;
        end if;
      else
        next_state(CTLR_ACT_WAIT)<=logic1;
        if (read_cmd_reg = '1') then
          read_cmd <= '1';
        end if;
      end if;
      
    elsif state(CTLR_ACT_WAIT) = '1' then  --issue NOP command, wait for rcd_end
      if (rcd_end = '1') then
        if(read_cmd_reg = '1') then
          next_state(CTLR_READ)<=logic1;
        else
          next_state(CTLR_WRITE)<=logic1;
        end if;
      else
        next_state(CTLR_ACT_WAIT)<=logic1;
        if(read_cmd_reg = '1') then
          read_cmd <= '1';
        end if;
      end if;
      
    elsif state(CTLR_READ) = '1' then
      if (cas_lat_end = '0') then
        next_state(CTLR_READ_WAIT)<=logic1;
      else
        next_state(CTLR_READ_DATA)<=logic1;
      end if;
      
    elsif state(CTLR_READ_WAIT) = '1' then
      if (cas_lat_end = '0') then
        next_state(CTLR_READ_WAIT)<=logic1;
      elsif (burst_2 = '1') then
        next_state(CTLR_IDLE)<=logic1;
      else
        next_state(CTLR_READ_DATA)<=logic1;
      end if;
      
    elsif state(CTLR_WRITE) = '1' then
      if (burst_2 = '1') then
        next_state(CTLR_IDLE)<=logic1;
      else
        next_state(CTLR_WRITE_DATA)<=logic1;
      end if;
      
    elsif state(CTLR_READ_DATA) = '1' then
      if (burst_8 = '1' and burst_end = '0') then
        next_state(CTLR_READ_DATA)<=logic1;
      else
        next_state(CTLR_IDLE)<=logic1;
      end if;
      
    elsif state(CTLR_WRITE_DATA) = '1' then
      if (burst_8 = '1' and burst_end = '0') then
        next_state(CTLR_WRITE_DATA)<=logic1;
      else
        next_state(CTLR_IDLE)<=logic1;
      end if;
      
    else
      next_state(CTLR_IDLE)<=logic1;
    end if;
  end process;

  -----------------------------------------------------------------------------
  --combinatorial next state and output logic ---------------------------------
  -----------------------------------------------------------------------------
  next_mrs_addr      <= u_cmd(SYS_LOAD_MR);
  next_row_addr      <= not(next_state(CTLR_PRECHARGE) or next_state(CTLR_READ)
                            or next_state(CTLR_WRITE));
  next_ld_cas_lat    <= not(state(CTLR_READ) or state(CTLR_READ_WAIT));
  next_ld_burst      <= not(state(CTLR_READ) or state(CTLR_WRITE) or data_read_write);
  next_ld_rcd        <= not(state(CTLR_ACT) or state(CTLR_ACT_WAIT));
  next_u_ref_ack     <= next_state(CTLR_REFRESH);
  next_ddr_rasb_o    <= not(state(CTLR_REFRESH) or state(CTLR_PRECHARGE)
                            or state(CTLR_LOAD_MR) or state(CTLR_ACT));
  next_ddr_casb_o    <= not(state(CTLR_REFRESH) or state(CTLR_LOAD_MR)
                            or state(CTLR_READ) or state(CTLR_WRITE));
  next_ddr_web_o     <= not(state(CTLR_PRECHARGE) or state(CTLR_LOAD_MR)
                            or state(CTLR_WRITE));
  next_ddr_write_en  <= (state(CTLR_WRITE) or state(CTLR_WRITE_DATA));
  next_ddr_read_en   <= state(CTLR_READ_WAIT) or state(CTLR_READ_DATA);
  process(clk)
  begin
    if (clk 'event and clk = '1') then
      next_ddr_read_en_P1 <= next_ddr_read_en;
    end if;
  end process;

  process (burst_8, next_ddr_read_en, next_ddr_read_en_P1)
  begin  -- process
    if burst_8 = '1' then
      next_u_data_valid_en <= next_ddr_read_en;
    else
      -- for burst of 4 or 2, need to delay u_data_valid_en
      next_u_data_valid_en <= next_ddr_read_en or next_ddr_read_en_P1;
    end if;
  end process;

  -----------------------------------------------------------------------------
  --output and state registers ------------------------------------------------
  -----------------------------------------------------------------------------
  --ddr control signals are generated on negative edge clock to guarantee DDR hold time
  process(clk,cntlr_reset1)
  begin
    if (cntlr_reset1 = '1') then
      ddr_rasb_o<= '1';
      ddr_casb_o<= '1';
      ddr_web_o <= '1';
    elsif (clk 'event and clk = '0') then  
      ddr_rasb_o       <= ddr_rasb_o_P1;
      ddr_casb_o       <= ddr_casb_o_P1;
      ddr_web_o        <= ddr_web_o_P1;          
    end if;
  end process;
  
  process(clk)
  begin
    if (clk 'event and clk = '1') then 
      ddr_rasb_o_P1    <= next_ddr_rasb_o;
      ddr_casb_o_P1    <= next_ddr_casb_o;
      ddr_web_o_P1     <= next_ddr_web_o;
    end if;
  end process;
  
  process(clk,cntlr_reset2)
  begin
    if (cntlr_reset2 = '1') then
      state(CTLR_IDLE)      <=logic1;
      state(CTLR_REFRESH)   <=logic0;
      state(CTLR_PRECHARGE) <=logic0;
      state(CTLR_LOAD_MR)   <=logic0;
      state(CTLR_ACT)       <=logic0;
      state(CTLR_ACT_WAIT)  <=logic0;
      state(CTLR_READ)      <=logic0;
      state(CTLR_WRITE)     <=logic0;
      state(CTLR_READ_WAIT) <=logic0;
      state(CTLR_READ_DATA) <=logic0;
      state(CTLR_WRITE_DATA)<=logic0;
      read_cmd_reg          <=logic0;
    elsif (clk 'event and clk = '1') then
      g0: for i in 0 to ENABLE_MSB loop
        ddr_read_en(i) <= next_ddr_read_en_P1;
      end loop;
      read_cmd_reg <= read_cmd;
      mrs_addr	 <= next_mrs_addr;
      row_addr	 <= next_row_addr;
      ld_cas_lat	 <= next_ld_cas_lat;
      ld_burst	 <= next_ld_burst;
      ld_rcd	 <= next_ld_rcd;
      u_ref_ack	 <= next_u_ref_ack;
      state         <= next_state;
      u_data_valid_en <= next_u_data_valid_en;
    end if;
  end process;
  
  --ddr_write_en is used to enable ddr_dq signals
  process(clk)
  begin
    if (clk 'event and clk ='1') then
      next_ddr_write_en_P1 <= next_ddr_write_en;
    end if;
  end process;
  
  process(clk2x)
  begin
    if (clk2x 'event and clk2x ='1') then
      g1: for i in 0 to ENABLE_MSB loop
        ddr_write_en(i) <= next_ddr_write_en_P1;
      end loop;
    end if;
  end process;

  --ddr_dqs_t is used to generate ddr_dqs signals
  process(clk)
  begin
    if (clk 'event and clk = '1') then
      ddr_dqs_t_tmp1 <= next_ddr_write_en;
      ddr_dqs_t_tmp2 <= ddr_dqs_t_tmp1;
    end if;
  end process;

  --create local lac_clk and also generate lac_clk_out
  lac_clk_in_inv <= not (lac_clk_in);
  clk2x_inv <= not (clk2x);
  LAC1: FD port map(D=>lac_clk_in_inv, C=>clk2x_inv, Q=>lac_clk_out);
  LAC2: FD port map(D=>lac_clk_in_inv, C=>clk2x_inv, Q=>lac_clk_wire);
  
  lac_clk <= lac_clk_wire after 1 ns;
  
  process(clk2x, cntlr_reset3)
  begin
    if (cntlr_reset3 ='1') then
      ddr_dqs_t <= '1';
    elsif (clk2x 'event and clk2x = '0') then
      if (lac_clk = '1') then
        ddr_dqs_t <= not (ddr_dqs_t_tmp1 or next_ddr_write_en);
      end if;
    end if;
  end process;
  
  o_ddr_ras: OBUF_SSTL2_I port map(I=>ddr_rasb_o, O=>ddr_rasb);
  o_ddr_cas: OBUF_SSTL2_I port map(I=>ddr_casb_o, O=>ddr_casb);
  o_ddr_web: OBUF_SSTL2_I port map(I=>ddr_web_o, O=>ddr_web);
   
end rtl;
