/******************************************************************************
 *
 *    File Name:  data_path.v
 *      Version:  2.5
 *         Date:  June 05, 2000
 * 
 *  Description:  generates ddr_dq and ddr_dqs during write cycles.  
 *                Receives ddr_dq during read cycle and generate u_data_o and u_data_valid
 * 
 *                Write:
 *                32-bit u_data_i is muxed into 16-bit ddr_dq with clk2x. 
 *                lac_clk selects between higher and lower data.
 *                tristate signal for ddr_dq (ddr_dq_t) is generated from ddr_write_en
 *                ddr_dqs is generated by a toggle flipflop and clocked by negedge of clk2x.
 *                this ensures dqs edges are at the center of dq
 * 
 *                Read:
 *                16-bit ddr_dq is clocked in by clk2x.  They are then assembled into
 *                32-bit data with higher bits registered by posedge of clk, and lower bits
 *                registered by negedge clk.
 * 
 *                All inputs and outputs to DDR SDRAM are registered in the IOB to guarantee
 *                minimum clock-to-out delay 
 *
 * 
 *       Author:  Jennnifer Tran
 *      Company:  Xilinx
 *
 *                Copyright (c) 1999 Xilinx, Inc.
 *                All rights reserved
 *
 *    DDR SDRAM:  64M (1M x 16Bit x 4Bank)
 * 
 *   Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
 *                WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
 *                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
 *                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
 *
 ******************************************************************************/

`timescale  100 ps / 10 ps
`include "../src/define.v"

module data_path (/*AUTOARG*/
   // Outputs
   u_data_o, ddr_dqs, u_data_valid, 
   // Inouts
   ddr_dq, 
   // Inputs
   u_data_i, ddr_write_en, ddr_dqs_t, ddr_read_en, clk2x, clk, 
   lac_clk_in, cas_lat_half, burst_8, u_data_valid_en
   );

   input [`U_DATA_MSB:0]     u_data_i;
   input [`ENABLE_MSB:0]     ddr_write_en;
   input 		     ddr_dqs_t;
   input [`ENABLE_MSB:0]     ddr_read_en;
   input 		     clk2x, clk, lac_clk_in;	
   input [1:0]		     cas_lat_half;
   input 		     burst_8;
   input 		     u_data_valid_en;
   inout [`DDR_DATA_MSB:0]   ddr_dq;
   output [`U_DATA_MSB:0]    u_data_o;
   output [3:0] 	     ddr_dqs;
   output 		     u_data_valid;

   /**************************************
    * pipeline/duplicate control signals *
    **************************************/
   //pipeline ddr_write_en
   reg [`ENABLE_MSB:0] 	     ddr_write_en_P1_5;
   always @(posedge clk) begin
	ddr_write_en_P1_5 <= ddr_write_en;
   end

   //duplicate tristate signal to reduce net delay
   wire [`ENABLE_MSB:0] ddr_dq_t_wire = ~(ddr_write_en_P1_5);
   
   //clk is used to select b/t lower and higher bits of write_data
   //and use clk2x to generate double data output
   //However, the clk signal cannot drive non clock inputs in Virtex, so
   //we need to use Logic Accessible Clock signal

   //pipeline Logic Accessible Clock signal to reduce delay
`include "../src/lac_clock.v"

    
   /***************
    * Write cycle *
    ***************/

   //generate ddr_dq
   reg [`U_DATA_MSB:0] write_data;

   //delay input data for 5 clk cycles
   reg [`U_DATA_MSB:0]  u_data_i_P1, u_data_i_P2, u_data_i_P3, u_data_i_P4, u_data_i_P5;
   always @(posedge clk) begin
      u_data_i_P1 <= u_data_i;
      u_data_i_P2 <= u_data_i_P1;
      u_data_i_P3 <= u_data_i_P2;
      u_data_i_P4 <= u_data_i_P3;
      u_data_i_P5 <= u_data_i_P4;
   end
   always @(negedge clk)
      write_data <= u_data_i_P5;

         
   //mux 128-bit write data into 16-bit ddr_dq   
   reg [`DDR_DATA_MSB:0]     ddr_dq_o, ddr_dq_o_tmp;
   reg [`DDR_DATA_MSB:0]    ddr_dq_t;
   always @(posedge clk2x) begin
`include  "../src/data_mux.v"
   end // always @ (posedge clk2x)

   always @(posedge clk2x) begin
      //additional reg to be pulled into IOB
      ddr_dq_o <= #18 ddr_dq_o_tmp;  //simulate clk-to-out delay on IOB flip-flop
   end
   

   //generate tristate signal for ddr_dq
   //ddr_dq_t f/f should be in IOB
   always @(posedge clk2x) begin
      ddr_dq_t[3:0]   <= #18 {4{ddr_dq_t_wire[0]}};
      ddr_dq_t[7:4]   <= #18 {4{ddr_dq_t_wire[1]}};
      ddr_dq_t[11:8]  <= #18 {4{ddr_dq_t_wire[2]}};
      ddr_dq_t[15:12] <= #18 {4{ddr_dq_t_wire[3]}};
      ddr_dq_t[19:16] <= #18 {4{ddr_dq_t_wire[4]}};
      ddr_dq_t[23:20] <= #18 {4{ddr_dq_t_wire[5]}};
      ddr_dq_t[27:24] <= #18 {4{ddr_dq_t_wire[6]}};
      ddr_dq_t[31:28] <= #18 {4{ddr_dq_t_wire[7]}};
      ddr_dq_t[35:32] <= #18 {4{ddr_dq_t_wire[8]}};
      ddr_dq_t[39:36] <= #18 {4{ddr_dq_t_wire[9]}};
      ddr_dq_t[43:40] <= #18 {4{ddr_dq_t_wire[10]}};
      ddr_dq_t[47:44] <= #18 {4{ddr_dq_t_wire[11]}};
      ddr_dq_t[51:48] <= #18 {4{ddr_dq_t_wire[12]}};
      ddr_dq_t[55:52] <= #18 {4{ddr_dq_t_wire[13]}};
      ddr_dq_t[59:56] <= #18 {4{ddr_dq_t_wire[14]}};
      ddr_dq_t[63:60] <= #18 {4{ddr_dq_t_wire[15]}};
    end

   //generate dqs
   //toggle f/f will be implemented in CLB
   reg 	     ddr_dqs_tog1, ddr_dqs_tog2;
   reg [3:0] ddr_dqs_o_tmp;
   always @(negedge clk2x) begin
      ddr_dqs_tog1 <= ddr_dqs_t;
      ddr_dqs_tog2 <= ddr_dqs_tog1;
   end
   always @(negedge clk2x) begin
      if (ddr_dqs_t)
	ddr_dqs_o_tmp <= #1 4'b0000;
      else 
	ddr_dqs_o_tmp <= #1 ~ddr_dqs_o_tmp;
   end
   reg [3:0] ddr_dqs_o_tmp_P1, ddr_dqs_o_tmp_P2;
   //delay ddr_dqs_o_tmp by 2 clk2x cycles
   always @(negedge clk2x) begin
      ddr_dqs_o_tmp_P1 <= ddr_dqs_o_tmp;
      ddr_dqs_o_tmp_P2 <= ddr_dqs_o_tmp_P1;
   end
   
   //the following f/fs will be in IOB
   reg [3:0] ddr_dqs_o_reg, ddr_dqs_t_P1, ddr_dqs_t_reg;
   always @(negedge clk2x)  begin
      ddr_dqs_o_reg <= #24 ddr_dqs_o_tmp_P2;  //simulate clk-to-out delay on IOB flip-flop
      ddr_dqs_t_P1 <= {4{ddr_dqs_t}};
      ddr_dqs_t_reg <= #24 ddr_dqs_t_P1;   //simulate clk-to-out delay on IOB flip-flop
   end
     
   OBUFT_SSTL2_II ddr_dqs0 (.I(ddr_dqs_o_reg[0]), .O(ddr_dqs[0]), .T(ddr_dqs_t_reg[0]));
   OBUFT_SSTL2_II ddr_dqs1 (.I(ddr_dqs_o_reg[1]), .O(ddr_dqs[1]), .T(ddr_dqs_t_reg[1]));
   OBUFT_SSTL2_II ddr_dqs2 (.I(ddr_dqs_o_reg[2]), .O(ddr_dqs[2]), .T(ddr_dqs_t_reg[2]));
   OBUFT_SSTL2_II ddr_dqs3 (.I(ddr_dqs_o_reg[3]), .O(ddr_dqs[3]), .T(ddr_dqs_t_reg[3]));


   /**************
    * READ cycle *
    **************/

   //pipeline read control signals
   reg u_data_valid_en_P1, u_data_valid_en_P2, u_data_valid_en_P3, u_data_valid_en_P4;
   always @(posedge clk) begin
      u_data_valid_en_P1 <= u_data_valid_en;
      u_data_valid_en_P2 <= u_data_valid_en_P1;
      u_data_valid_en_P3 <= u_data_valid_en_P2;
      u_data_valid_en_P4 <= u_data_valid_en_P3;
   end
 
   // read in data
   wire [`DDR_DATA_MSB:0]    ddr_dq_i;
   reg [`DDR_DATA_MSB:0]     read_data;
   always @(posedge clk2x) begin
     //register 64-bit data from ddr sdram
      if (ddr_read_en[0]) read_data[3:0] <= ddr_dq_i[3:0];
      if (ddr_read_en[1]) read_data[7:4] <= ddr_dq_i[7:4];
      if (ddr_read_en[2]) read_data[11:8] <= ddr_dq_i[11:8];
      if (ddr_read_en[3]) read_data[15:12] <= ddr_dq_i[15:12];
      if (ddr_read_en[4]) read_data[19:16] <= ddr_dq_i[19:16];
      if (ddr_read_en[5]) read_data[23:20] <= ddr_dq_i[23:20];
      if (ddr_read_en[6]) read_data[27:24] <= ddr_dq_i[27:24];
      if (ddr_read_en[7]) read_data[31:28] <= ddr_dq_i[31:28];
      if (ddr_read_en[8]) read_data[35:32] <= ddr_dq_i[35:32];
      if (ddr_read_en[9]) read_data[39:36] <= ddr_dq_i[39:36];
      if (ddr_read_en[10]) read_data[43:40] <= ddr_dq_i[43:40];
      if (ddr_read_en[11]) read_data[47:44] <= ddr_dq_i[47:44];
      if (ddr_read_en[12]) read_data[51:48] <= ddr_dq_i[51:48];
      if (ddr_read_en[13]) read_data[55:52] <= ddr_dq_i[55:52];
      if (ddr_read_en[14]) read_data[59:56] <= ddr_dq_i[59:56];
      if (ddr_read_en[15]) read_data[63:60] <= ddr_dq_i[63:60];
    end

   //join 64-bit DDR read_data into 128-bit u_ad
   wire [`DDR_DATA_MSB:0] #1 read_data_dly = read_data; //simulate clock-to-out delay
   reg  [`U_DATA_MSB:0]  u_data_o_20, u_data_o_20_tmp, u_data_o_20_P1;
   reg  [`U_DATA_MSB:0]  u_data_o_25, u_data_o_25_tmp, u_data_o_25_P1;
   wire [`U_DATA_MSB:0] u_data_o;

   //CAS latency 2.5
   always @(negedge clk) u_data_o_25_tmp[`U_DATA_MSB:`DDR_DATA_MSB+1] <= read_data_dly;
   always @(posedge clk) u_data_o_25_tmp[`DDR_DATA_MSB:0]  <= read_data_dly;
   always @(negedge clk) u_data_o_25_P1 <= u_data_o_25_tmp;
   always @(posedge clk) u_data_o_25 <= u_data_o_25_P1;

   // CAS latency 2
   always @(posedge clk) u_data_o_20_tmp[`U_DATA_MSB:`DDR_DATA_MSB+1]  <= read_data_dly;
   always @(negedge clk) u_data_o_20_tmp[`DDR_DATA_MSB:0]  <= read_data_dly;
   always @(posedge clk) u_data_o_20 <= u_data_o_20_tmp;

   assign u_data_o = cas_lat_half[0]? u_data_o_25 : u_data_o_20;
	  
  //generate u_data_valid   
   reg u_data_valid_20, u_data_valid_25;
   wire u_data_valid;
   always @(posedge clk)
     if (u_data_valid_en_P4 && u_data_valid_en_P1) u_data_valid_25 <= 1;
     else       u_data_valid_25 <= 0;
   
    always @(posedge clk)
      if (u_data_valid_en_P3 && u_data_valid_en) u_data_valid_20 <= 1;
      else       u_data_valid_20 <= 0;

   assign u_data_valid = cas_lat_half[1] ? u_data_valid_25 : u_data_valid_20;
  

   //IO buffers to interface with DDR SDRAM *
   wire [`DDR_DATA_MSB:0] ddr_dq_i_wire;
   assign    ddr_dq_i = ddr_dq_i_wire;

   ddr_dq_io_64 ddr_dq_io (.ddr_dq_i(ddr_dq_i_wire), .ddr_dq(ddr_dq), .ddr_dq_o(ddr_dq_o), .ddr_dq_t(ddr_dq_t));
   
endmodule 




